#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__all__ = [
    "plausible_url",
    "plausible_browse_url",
    "sanitize_url",
    "is_gitlab_site",
    "browse_url_from_repo_url",
]

import http.client
import logging
import re
from typing import Optional, Union, List, Tuple

import socket
import urllib
from urllib.parse import urlparse, urlunparse, ParseResult, parse_qs


from . import (
    _load_json_url,
    InvalidUrl,
    UrlUnverifiable,
)


KNOWN_GITLAB_SITES = [
    "salsa.debian.org",
    "invent.kde.org",
    "0xacab.org",
]


KNOWN_HOSTING_SITES = [
    'code.launchpad.net', 'github.com', 'launchpad.net', 'git.openstack.org']


logger = logging.getLogger(__name__)


def plausible_browse_url(url: str) -> bool:
    return url.startswith("https://") or url.startswith("http://")


def plausible_url(url: str) -> bool:
    return ":" in url


def unsplit_vcs_url(
    repo_url: str, branch: Optional[str] = None, subpath: Optional[str] = None
) -> str:
    """Unsplit a Debian VCS URL.

    Args:
      repo_url: Repository URL
      branch: Branch name
      subpath: Subpath in the tree
    Returns: full URL
    """
    url = repo_url
    if branch:
        url = "{} -b {}".format(url, branch)
    if subpath:
        url = "{} [{}]".format(url, subpath)
    return url


def probe_gitlab_host(hostname: str) -> bool:
    import json
    try:
        _load_json_url("https://%s/api/v4/version" % hostname)
    except urllib.error.HTTPError as e:
        if e.code == 401:
            try:
                if json.loads(e.read()) == {"message": "401 Unauthorized"}:
                    return True
            except json.JSONDecodeError:
                return False
        return False
    except UnicodeDecodeError:
        return False
    except json.JSONDecodeError:
        return False
    except (socket.timeout, urllib.error.URLError):
        # Probably not?
        return False
    except http.client.RemoteDisconnected:
        return False
    return False


def is_gitlab_site(hostname: str, net_access: bool = False) -> bool:
    if hostname is None:
        return False
    if hostname in KNOWN_GITLAB_SITES:
        return True
    if hostname.startswith("gitlab."):
        return True
    if net_access:
        return probe_gitlab_host(hostname)
    return False


def browse_url_from_repo_url(  # noqa: C901
        url: str, *, branch: Optional[str] = None,
        subpath: Optional[str] = None) -> Optional[str]:
    if isinstance(url, list):
        return None
    parsed_url = urlparse(url)
    if parsed_url.netloc == "github.com":
        path = "/".join(parsed_url.path.split("/")[:3])
        if path.endswith(".git"):
            path = path[:-4]
        if subpath is not None or branch is not None:
            path += "/tree/%s" % (branch or "HEAD")
        if subpath is not None:
            path += "/" + subpath
        return urlunparse(("https", "github.com", path, None, None, None))
    elif parsed_url.hostname == 'gopkg.in':
        els = parsed_url.path.split("/")[:3]
        if len(els) != 2:
            return None
        try:
            els[-1], version = els[-1].split('.v', 1)
        except ValueError:
            els[-1] = els[-1]
            version = "HEAD"
        els.extend(['tree', version])
        path = "/".join(els)
        if subpath is not None:
            path += "/" + subpath
        return urlunparse(("https", "github.com", path, None, None, None))
    elif parsed_url.netloc in ("code.launchpad.net", "launchpad.net"):
        if subpath is not None:
            path = parsed_url.path + "/view/head:/" + subpath
            return urlunparse(
                (
                    "https",
                    "bazaar.launchpad.net",
                    path,
                    parsed_url.query,
                    parsed_url.params,
                    parsed_url.fragment,
                )
            )
        else:
            return urlunparse(
                (
                    "https",
                    "code.launchpad.net",
                    parsed_url.path,
                    parsed_url.query,
                    parsed_url.params,
                    parsed_url.fragment,
                )
            )
    elif parsed_url.netloc == "svn.apache.org":
        path_elements = parsed_url.path.strip("/").split("/")
        if path_elements[:2] != ["repos", "asf"]:
            return None
        path_elements.pop(0)
        path_elements[0] = "viewvc"
        if subpath is not None:
            path_elements.append(subpath)
        return urlunparse(
            ("https", parsed_url.netloc, "/".join(path_elements), None, None, None)
        )
    elif parsed_url.hostname in ("git.savannah.gnu.org", "git.sv.gnu.org"):
        path_elements = parsed_url.path.strip("/").split("/")
        if parsed_url.scheme == "https" and path_elements[0] == "git":
            path_elements.pop(0)
        # Why cgit and not gitweb?
        path_elements.insert(0, "cgit")
        if subpath is not None:
            path_elements.append("tree")
            path_elements.append(subpath)
        return urlunparse(
            ("https", parsed_url.netloc, "/".join(path_elements), None, None, None)
        )
    elif is_gitlab_site(parsed_url.netloc):
        path = parsed_url.path
        if path.endswith(".git"):
            path = path[:-4]
        if subpath is not None:
            path += "/-/blob/HEAD/" + subpath
        return urlunparse(("https", parsed_url.netloc, path, None, None, None))

    return None


SECURE_SCHEMES = ["https", "git+ssh", "bzr+ssh", "hg+ssh", "ssh", "svn+ssh"]


def try_open_branch(url: str, branch_name: Optional[str] = None):
    import breezy.ui
    from breezy.controldir import ControlDir

    old_ui = breezy.ui.ui_factory
    breezy.ui.ui_factory = breezy.ui.SilentUIFactory()
    try:
        c = ControlDir.open(url)
        b = c.open_branch(name=branch_name)
        b.last_revision()
        return b
    except Exception:
        # TODO(jelmer): Catch more specific exceptions?
        return None
    finally:
        breezy.ui.ui_factory = old_ui


def find_secure_repo_url(
    url: str, branch: Optional[str] = None, net_access: bool = True
) -> Optional[str]:
    parsed_repo_url = urlparse(url)
    if parsed_repo_url.scheme in SECURE_SCHEMES:
        return url

    # Sites we know to be available over https
    if (parsed_repo_url.hostname
        and (
            is_gitlab_site(parsed_repo_url.hostname, net_access)
            or parsed_repo_url.hostname in [
                "github.com",
                "git.launchpad.net",
                "bazaar.launchpad.net",
                "code.launchpad.net",
            ])):
        parsed_repo_url = parsed_repo_url._replace(scheme="https")

    if parsed_repo_url.scheme == "lp":
        parsed_repo_url = parsed_repo_url._replace(
            scheme="https", netloc="code.launchpad.net"
        )

    if parsed_repo_url.hostname in ("git.savannah.gnu.org", "git.sv.gnu.org"):
        if parsed_repo_url.scheme == "http":
            parsed_repo_url = parsed_repo_url._replace(scheme="https")
        else:
            parsed_repo_url = parsed_repo_url._replace(
                scheme="https", path="/git" + parsed_repo_url.path
            )

    if net_access:
        secure_repo_url = parsed_repo_url._replace(scheme="https")
        insecure_branch = try_open_branch(url, branch)
        secure_branch = try_open_branch(urlunparse(secure_repo_url), branch)
        if secure_branch:
            if (
                not insecure_branch
                or secure_branch.last_revision() == insecure_branch.last_revision()
            ):
                parsed_repo_url = secure_repo_url

    if parsed_repo_url.scheme in SECURE_SCHEMES:
        return urlunparse(parsed_repo_url)

    # Can't find a secure URI :(
    return None


def canonical_git_repo_url(repo_url: str) -> str:
    parsed_url = urlparse(repo_url)
    if is_gitlab_site(parsed_url.netloc) or parsed_url.netloc in ["github.com"]:
        if not parsed_url.path.rstrip("/").endswith(".git"):
            parsed_url = parsed_url._replace(path=parsed_url.path.rstrip("/") + ".git")
        return urlunparse(parsed_url)
    return repo_url


def find_public_repo_url(repo_url: str) -> Optional[str]:
    parsed = urlparse(repo_url)
    if not parsed.scheme and not parsed.hostname and ':' in parsed.path:
        m = re.match('^(?P<user>[^@:/]+@)?(?P<host>[^/:]+):(?P<path>.*)$',
                     repo_url)
        if m:
            host = m.group('host')
            path = m.group('path')
            if host == 'github.com' or is_gitlab_site(host):
                return urlunparse(("https", "github.com", path, None, None, None))

    parsed = urlparse(repo_url)
    revised_url = None
    if parsed.hostname == "github.com":
        if parsed.scheme in ("https", "http", "git"):
            return repo_url
        revised_url = urlunparse(("https", "github.com", parsed.path, None, None, None))
    if parsed.hostname and is_gitlab_site(parsed.hostname):
        # Not sure if gitlab even support plain http?
        if parsed.scheme in ("https", "http"):
            return repo_url
        if parsed.scheme == "ssh":
            revised_url = urlunparse(
                ("https", parsed.hostname, parsed.path, None, None, None)
            )
    if parsed.hostname in (
        "code.launchpad.net",
        "bazaar.launchpad.net",
        "git.launchpad.net",
    ):
        if parsed.scheme.startswith("http") or parsed.scheme == "lp":
            return repo_url
        if parsed.scheme in ("ssh", "bzr+ssh"):
            revised_url = urlunparse(
                ("https", parsed.hostname, parsed.path, None, None, None)
            )

    if revised_url:
        return revised_url

    return None


def fixup_rcp_style_git_repo_url(url: str) -> str:
    try:
        from breezy.location import rcp_location_to_url
    except ModuleNotFoundError:
        return url

    try:
        repo_url = rcp_location_to_url(url)
    except ValueError:
        return url
    return repo_url


def drop_vcs_in_scheme(url: str) -> str:
    if url.startswith("git+http:") or url.startswith("git+https:"):
        url = url[4:]
    if url.startswith("hg+https:") or url.startswith("hg+http"):
        url = url[3:]
    if url.startswith("bzr+lp:") or url.startswith("bzr+http"):
        url = url.split("+", 1)[1]
    return url


def fix_path_in_port(
    parsed: ParseResult, branch: Optional[str], subpath: Optional[str]
):
    if ":" not in parsed.netloc or parsed.netloc.endswith("]"):
        return None, None, None
    host, port = parsed.netloc.rsplit(":", 1)
    if host.split("@")[-1] not in (KNOWN_GITLAB_SITES + ["github.com"]):
        return None, None, None
    if not port or port.isdigit():
        return None, None, None
    return (
        parsed._replace(path="{}/{}".format(port, parsed.path.lstrip("/")), netloc=host),
        branch,
        subpath,
    )


def fix_gitlab_scheme(parsed, branch, subpath: Optional[str]):
    if is_gitlab_site(parsed.hostname):
        return parsed._replace(scheme="https"), branch, subpath
    return None, None, None


def fix_github_scheme(parsed, branch, subpath: Optional[str]):
    # GitHub no longer supports the git:// scheme
    if parsed.hostname == 'github.com' and parsed.scheme == 'git':
        return parsed._replace(scheme='https'), branch, subpath
    return None, None, None


def fix_salsa_cgit_url(parsed, branch, subpath):
    if parsed.hostname == "salsa.debian.org" and parsed.path.startswith("/cgit/"):
        return parsed._replace(path=parsed.path[5:]), branch, subpath
    return None, None, None


def fix_gitlab_tree_in_url(parsed, branch, subpath):
    if is_gitlab_site(parsed.hostname):
        parts = parsed.path.split("/")
        if len(parts) >= 5 and parts[3] == "tree":
            branch = "/".join(parts[4:])
            return parsed._replace(path="/".join(parts[:3])), branch, subpath
    return None, None, None


def fix_double_slash(parsed, branch, subpath):
    if parsed.path.startswith("//"):
        return parsed._replace(path=parsed.path[1:]), branch, subpath
    return None, None, None


def fix_extra_colon(parsed, branch, subpath):
    return parsed._replace(netloc=parsed.netloc.rstrip(":")), branch, subpath


def drop_git_username(parsed, branch, subpath):
    if parsed.hostname not in ("salsa.debian.org", "github.com"):
        return None, None, None
    if parsed.scheme not in ("git", "http", "https"):
        return None, None, None
    if parsed.username == "git" and parsed.netloc.startswith("git@"):
        return parsed._replace(netloc=parsed.netloc[4:]), branch, subpath
    return None, None, None


def fix_branch_argument(parsed, branch, subpath):
    if parsed.hostname == "github.com":
        # TODO(jelmer): Handle gitlab sites too?
        path_elements = parsed.path.strip("/").split("/")
        if len(path_elements) > 2 and path_elements[2] == "tree":
            return (
                parsed._replace(path="/".join(path_elements[:2])),
                "/".join(path_elements[3:]),
                subpath,
            )
    return None, None, None


def fix_git_gnome_org_url(parsed, branch, subpath):
    if parsed.netloc == "git.gnome.org":
        if parsed.path.startswith("/browse"):
            path = parsed.path[7:]
        else:
            path = parsed.path
        parsed = parsed._replace(
            netloc="gitlab.gnome.org", scheme="https", path="/GNOME" + path
        )
        return parsed, branch, subpath
    return None, None, None


def fix_anongit_url(parsed, branch, subpath):
    if parsed.netloc == "anongit.kde.org" and parsed.scheme == "git":
        parsed = parsed._replace(scheme="https")
        return parsed, branch, subpath
    return None, None, None


def fix_freedesktop_org_url(
    parsed: ParseResult, branch: Optional[str], subpath: Optional[str]
):
    if parsed.netloc == "anongit.freedesktop.org":
        path = parsed.path
        if path.startswith("/git/"):
            path = path[len("/git") :]
        parsed = parsed._replace(
            netloc="gitlab.freedesktop.org", scheme="https", path=path
        )
        return parsed, branch, subpath
    return None, None, None


FIXERS = [
    fix_path_in_port,
    fix_gitlab_scheme,
    fix_github_scheme,
    fix_salsa_cgit_url,
    fix_gitlab_tree_in_url,
    fix_double_slash,
    fix_extra_colon,
    drop_git_username,
    fix_branch_argument,
    fix_git_gnome_org_url,
    fix_anongit_url,
    fix_freedesktop_org_url,
]


def fixup_broken_git_details(
    repo_url: str, branch: Optional[str], subpath: Optional[str]
) -> Tuple[str, Optional[str], Optional[str]]:
    """Attempt to fix up broken Git URLs.

    A common misspelling is to add an extra ":" after the hostname
    """
    parsed = urlparse(repo_url)
    changed = False
    for fn in FIXERS:
        newparsed, newbranch, newsubpath = fn(parsed, branch, subpath)
        if newparsed:
            changed = True
            parsed = newparsed
            branch = newbranch
            subpath = newsubpath

    if changed:
        return urlunparse(parsed), branch, subpath

    return repo_url, branch, subpath


def convert_cvs_list_to_str(urls):
    if not isinstance(urls, list):
        return urls
    if urls[0].startswith(":extssh:") or urls[0].startswith(":pserver:"):
        from breezy.location import cvs_to_url
        return cvs_to_url(urls[0]) + "#" + urls[1]
    return urls


SANITIZERS = [
    convert_cvs_list_to_str,
    drop_vcs_in_scheme,
    lambda url: fixup_broken_git_details(url, None, None)[0],
    fixup_rcp_style_git_repo_url,
    lambda url: find_public_repo_url(url) or url,
    canonical_git_repo_url,
    lambda url: find_secure_repo_url(url, net_access=False) or url,
]


def sanitize_url(url: Union[str, List[str]]) -> str:
    if isinstance(url, str):
        url = url.strip()
    for sanitizer in SANITIZERS:
        url = sanitizer(url)
    return url  # type: ignore


def guess_repo_from_url(url, net_access=False):  # noqa: C901
    if isinstance(url, list):
        return None
    parsed_url = urlparse(url)
    path_elements = parsed_url.path.strip('/').split('/')
    if parsed_url.netloc == 'github.com':
        if len(path_elements) < 2:
            return None
        return ('https://github.com'
                + '/'.join(parsed_url.path.split('/')[:3]))
    if parsed_url.netloc == 'travis-ci.org':
        return ('https://github.com/'
                + '/'.join(path_elements[:3]))
    if (parsed_url.netloc == 'coveralls.io'
            and parsed_url.path.startswith('/r/')):
        return ('https://github.com/'
                + '/'.join(path_elements[1:4]))
    if parsed_url.netloc == 'launchpad.net':
        return 'https://code.launchpad.net/%s' % (
            parsed_url.path.strip('/').split('/')[0])
    if parsed_url.netloc == 'git.savannah.gnu.org':
        if len(path_elements) != 2 or path_elements[0] != 'git':
            return None
        return url
    if parsed_url.netloc in ('freedesktop.org', 'www.freedesktop.org'):
        if len(path_elements) >= 2 and path_elements[0] == 'software':
            return 'https://github.com/freedesktop/%s' % path_elements[1]
        if len(path_elements) >= 3 and path_elements[:2] == [
                'wiki', 'Software']:
            return 'https://github.com/freedesktop/%s.git' % path_elements[2]
    if parsed_url.netloc == 'download.gnome.org':
        if len(path_elements) >= 2 and path_elements[0] == 'sources':
            return 'https://gitlab.gnome.org/GNOME/%s.git' % path_elements[1]
    if parsed_url.netloc == 'download.kde.org':
        if len(path_elements) >= 2 and path_elements[0] in (
                'stable', 'unstable'):
            return 'https://anongit.kde.org/%s.git' % path_elements[1]
    if parsed_url.netloc == 'ftp.gnome.org':
        if (len(path_elements) >= 4 and [
            e.lower() for e in path_elements[:3]] == [
                'pub', 'gnome', 'sources']):
            return 'https://gitlab.gnome.org/GNOME/%s.git' % path_elements[3]
    if parsed_url.netloc == 'sourceforge.net':
        if (len(path_elements) >= 4 and path_elements[0] == 'p'
                and path_elements[3] == 'ci'):
            return 'https://sourceforge.net/p/{}/{}'.format(
                path_elements[1], path_elements[2])
    if parsed_url.netloc == 'www.apache.org':
        if len(path_elements) > 2 and path_elements[0] == 'dist':
            return 'https://svn.apache.org/repos/asf/{}/{}'.format(
                path_elements[1], path_elements[2])
    if parsed_url.netloc == 'bitbucket.org':
        if len(path_elements) >= 2:
            return 'https://bitbucket.org/{}/{}'.format(
                path_elements[0], path_elements[1])
    if parsed_url.netloc == 'ftp.gnu.org':
        if len(path_elements) >= 2 and path_elements[0] == 'gnu':
            return 'https://git.savannah.gnu.org/git/%s.git' % (
                path_elements[1])
        return None
    if parsed_url.netloc == 'download.savannah.gnu.org':
        if len(path_elements) >= 2 and path_elements[0] == 'releases':
            return 'https://git.savannah.gnu.org/git/%s.git' % (
                path_elements[1])
        return None
    if is_gitlab_site(parsed_url.hostname, net_access):
        if parsed_url.path.strip('/').count('/') < 1:
            return None
        parts = parsed_url.path.split('/')
        if '-' in parts:
            parts = parts[:parts.index('-')]
        if 'tags' in parts:
            parts = parts[:parts.index('tags')]
        if 'blob' in parts:
            parts = parts[:parts.index('blob')]
        return urlunparse(
            parsed_url._replace(path='/'.join(parts), query=''))
    if parsed_url.hostname == 'git.php.net':
        if parsed_url.path.startswith('/repository/'):
            return url
        if not parsed_url.path.strip('/'):
            qs = parse_qs(parsed_url.query)
            if 'p' in qs:
                return urlunparse(parsed_url._replace(
                    path='/repository/' + qs['p'][0], query=''))
    if parsed_url.netloc in KNOWN_HOSTING_SITES:
        return url
    # Maybe it's already pointing at a VCS repo?
    if parsed_url.netloc.startswith('svn.'):
        # 'svn' subdomains are often used for hosting SVN repositories.
        return url
    if net_access:
        try:
            return check_repository_url_canonical(url)
        except (UrlUnverifiable, InvalidUrl):
            return None
    return None


def check_repository_url_canonical(
        url: str, version: Optional[str] = None) -> str:
    parsed_url = urlparse(url)
    if parsed_url.netloc == 'github.com':
        path_elements = parsed_url.path.strip('/').split('/')
        if len(path_elements) < 2:
            raise InvalidUrl(
                url, "GitHub URL with less than 2 path elements")

        if path_elements[0] == 'sponsors':
            raise InvalidUrl(
                url, "GitHub sponsors URL")
        if path_elements[1].endswith('.git'):
            path_elements[1] = path_elements[1][:-4]
        api_url = 'https://api.github.com/repos/{}/{}'.format(
            path_elements[0], path_elements[1])
        try:
            data = _load_json_url(api_url)
        except urllib.error.HTTPError as e:
            if e.code == 404:
                raise InvalidUrl(
                    url, "API URL %s does not exist" % api_url) from e
            elif e.code == 403:
                # Probably rate limited
                logger.warning(
                    'Unable to verify bug database URL %s: %s',
                    url, e.reason)
                raise UrlUnverifiable(url, "GitHub URL rate-limited") from e
            else:
                raise
        else:
            if data.get('archived', False):
                raise InvalidUrl(url, "Repository is archived")
            if data['description']:
                if data['description'].startswith('Moved to '):
                    return check_repository_url_canonical(
                        data['description'][len('Moved to '):],
                        version=version)
                if 'has moved' in data['description']:
                    raise InvalidUrl(url, "repository has moved")
                if data['description'].startswith('Mirror of '):
                    return check_repository_url_canonical(
                        data['description'][len('Mirror of '):],
                        version=version)
            homepage = data.get('homepage')
            if homepage and is_gitlab_site(homepage):
                raise InvalidUrl(url, 'homepage is on GitLab: %s' % homepage)
            # TODO(jelmer): Look at the contents of the repository; if it
            # contains just a single README file with < 10 lines, assume
            # the worst.
            # return data['clone_url']
            url = data['clone_url']
    is_valid = probe_upstream_branch_url(url, version=version)
    if is_valid is None:
        raise UrlUnverifiable(url, "unable to probe")
    if is_valid:
        return url
    raise InvalidUrl(url, "unable to successfully probe URL")


def probe_upstream_branch_url(
        url: str, version: Optional[str] = None) -> Optional[bool]:
    parsed = urlparse(url)
    if parsed.scheme in ('git+ssh', 'ssh', 'bzr+ssh'):
        # Let's not probe anything possibly non-public.
        return None
    if parsed.hostname == 'github.com':
        path = parsed.path
        if path.endswith('.git'):
            path = path[:-4]
        api_url = ('https://api.github.com/repos/%s/tags'
                   % path.strip('/'))
        try:
            data = _load_json_url(api_url)
        except urllib.error.HTTPError as e:
            if e.code == 404:
                return False
            elif e.code == 403:
                logger.warning('Rate-limited by GitHub')
                return None
            else:
                raise
        else:
            tag_names = [k['name'] for k in data]
            if _version_in_tags(version, tag_names):
                return True
            return False
    else:
        import breezy.ui
        from breezy.branch import Branch
        import breezy.bzr
        import breezy.git
        old_ui = breezy.ui.ui_factory
        breezy.ui.ui_factory = breezy.ui.SilentUIFactory()
        try:
            b = Branch.open(url)
            b.last_revision()
            if version is not None:
                version = version.split('+git')[0]
                tag_names = b.tags.get_tag_dict().keys()
                if not tag_names:
                    # Uhm, hmm
                    return True
                if _version_in_tags(version, tag_names):
                    return True
                return False
            else:
                return True
        except Exception as e:
            logger.debug(
                'Error accessing %s: %s', url, e)
            # TODO(jelmer): Catch more specific exceptions?
            return False
        finally:
            breezy.ui.ui_factory = old_ui


def _version_in_tags(version, tag_names):
    if version in tag_names:
        return True
    if 'v%s' % version in tag_names:
        return True
    if 'release/%s' % version in tag_names:
        return True
    if version.replace('.', '_') in tag_names:
        return True
    for tag_name in tag_names:
        if tag_name.endswith('_' + version):
            return True
        if tag_name.endswith('-' + version):
            return True
        if tag_name.endswith('_%s' % version.replace('.', '_')):
            return True
    return False



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for upstream_ontologist."""

import os
import shutil
import tempfile
from unittest import (
    TestCase,
)


from upstream_ontologist import (
    UpstreamDatum,
    Person,
    min_certainty,
    certainty_to_confidence,
    confidence_to_certainty,
    certainty_sufficient,
)
from upstream_ontologist.guess import (
    guess_repo_from_url,
    guess_from_package_json,
    guess_from_debian_watch,
    guess_from_r_description,
    bug_database_url_from_bug_submit_url,
    url_from_git_clone_command,
    url_from_fossil_clone_command,
)


class TestCaseInTempDir(TestCase):
    def setUp(self):
        super().setUp()
        self.testdir = tempfile.mkdtemp()
        os.chdir(self.testdir)
        self.addCleanup(shutil.rmtree, self.testdir)


class GuessFromDebianWatchTests(TestCaseInTempDir):

    def setUp(self):
        super().setUp()
        try:
            import pcre  # noqa: F401
        except ModuleNotFoundError:
            self.skipTest('pcre not available')

    def test_empty(self):
        with open("watch", "w") as f:
            f.write(
                """\
# Blah
"""
            )
        self.assertEqual([], list(guess_from_debian_watch("watch", False)))

    def test_simple(self):
        with open("watch", "w") as f:
            f.write(
                """\
version=4
https://github.com/jelmer/dulwich/tags/dulwich-(.*).tar.gz
"""
            )
        self.assertEqual(
            [
                UpstreamDatum(
                    "Repository", "https://github.com/jelmer/dulwich", "likely", "watch"
                )
            ],
            list(guess_from_debian_watch("watch", False)),
        )


class GuessFromPackageJsonTests(TestCaseInTempDir):
    def test_simple(self):
        with open("package.json", "w") as f:
            f.write(
                """\
{
  "name": "autosize",
  "version": "4.0.2",
  "author": {
    "name": "Jack Moore",
    "url": "http://www.jacklmoore.com",
    "email": "hello@jacklmoore.com"
  },
  "main": "dist/autosize.js",
  "license": "MIT",
  "homepage": "http://www.jacklmoore.com/autosize",
  "demo": "http://www.jacklmoore.com/autosize",
  "repository": {
    "type": "git",
    "url": "http://github.com/jackmoore/autosize.git"
  }
}
"""
            )
        self.assertEqual(
            [
                UpstreamDatum("Name", "autosize", "certain"),
                UpstreamDatum(
                    "Homepage", "http://www.jacklmoore.com/autosize", "certain"
                ),
                UpstreamDatum("X-License", "MIT", "certain", None),
                UpstreamDatum("X-Version", "4.0.2", "certain"),
                UpstreamDatum(
                    "Repository", "http://github.com/jackmoore/autosize.git", "certain"
                ),
                UpstreamDatum(
                    'X-Author', [Person(
                        name="Jack Moore",
                        url="http://www.jacklmoore.com", 
                        email="hello@jacklmoore.com")], 'confident')
            ],
            list(guess_from_package_json("package.json", False)),
        )

    def test_dummy(self):
        with open("package.json", "w") as f:
            f.write(
                """\
{
  "name": "mozillaeslintsetup",
  "description": "This package file is for setup of ESLint.",
  "repository": {},
  "license": "MPL-2.0",
  "dependencies": {
    "eslint": "4.18.1",
    "eslint-plugin-html": "4.0.2",
    "eslint-plugin-mozilla": "file:tools/lint/eslint/eslint-plugin-mozilla",
    "eslint-plugin-no-unsanitized": "2.0.2",
    "eslint-plugin-react": "7.1.0",
    "eslint-plugin-spidermonkey-js":
        "file:tools/lint/eslint/eslint-plugin-spidermonkey-js"
  },
  "devDependencies": {}
}
"""
            )
        self.assertEqual(
            [
                UpstreamDatum("Name", "mozillaeslintsetup", "certain"),
                UpstreamDatum(
                    "X-Summary",
                    "This package file is for setup of ESLint.",
                    "certain",
                    None,
                ),
                UpstreamDatum("X-License", "MPL-2.0", "certain", None),
            ],
            list(guess_from_package_json("package.json", False)),
        )


class GuessFromRDescriptionTests(TestCaseInTempDir):
    def test_read(self):
        with open("DESCRIPTION", "w") as f:
            f.write(
                """\
Package: crul
Title: HTTP Client
Description: A simple HTTP client, with tools for making HTTP requests,
    and mocking HTTP requests. The package is built on R6, and takes
    inspiration from Ruby's 'faraday' gem (<https://rubygems.org/gems/faraday>)
    The package name is a play on curl, the widely used command line tool
    for HTTP, and this package is built on top of the R package 'curl', an
    interface to 'libcurl' (<https://curl.haxx.se/libcurl>).
Version: 0.8.4
License: MIT + file LICENSE
Authors@R: c(
    person("Scott", "Chamberlain", role = c("aut", "cre"),
    email = "myrmecocystus@gmail.com",
    comment = c(ORCID = "0000-0003-1444-9135"))
    )
URL: https://github.com/ropensci/crul (devel)
        https://ropenscilabs.github.io/http-testing-book/ (user manual)
        https://www.example.com/crul (homepage)
BugReports: https://github.com/ropensci/crul/issues
Encoding: UTF-8
Language: en-US
Imports: curl (>= 3.3), R6 (>= 2.2.0), urltools (>= 1.6.0), httpcode
        (>= 0.2.0), jsonlite, mime
Suggests: testthat, fauxpas (>= 0.1.0), webmockr (>= 0.1.0), knitr
VignetteBuilder: knitr
RoxygenNote: 6.1.1
X-schema.org-applicationCategory: Web
X-schema.org-keywords: http, https, API, web-services, curl, download,
        libcurl, async, mocking, caching
X-schema.org-isPartOf: https://ropensci.org
NeedsCompilation: no
Packaged: 2019-08-02 19:58:21 UTC; sckott
Author: Scott Chamberlain [aut, cre] (<https://orcid.org/0000-0003-1444-9135>)
Maintainer: Scott Chamberlain <myrmecocystus@gmail.com>
Repository: CRAN
Date/Publication: 2019-08-02 20:30:02 UTC
"""
            )
        ret = guess_from_r_description("DESCRIPTION", True)
        self.assertEqual(
            list(ret),
            [
                UpstreamDatum("Name", "crul", "certain"),
                UpstreamDatum("Archive", "CRAN", "certain"),
                UpstreamDatum(
                    "Bug-Database", "https://github.com/ropensci/crul/issues", "certain"
                ),
                UpstreamDatum('X-Version', '0.8.4', 'certain'),
                UpstreamDatum('X-License', 'MIT + file LICENSE', 'certain'),
                UpstreamDatum('X-Summary', 'HTTP Client', 'certain'),
                UpstreamDatum('X-Description', """\
A simple HTTP client, with tools for making HTTP requests,
and mocking HTTP requests. The package is built on R6, and takes
inspiration from Ruby's 'faraday' gem (<https://rubygems.org/gems/faraday>)
The package name is a play on curl, the widely used command line tool
for HTTP, and this package is built on top of the R package 'curl', an
interface to 'libcurl' (<https://curl.haxx.se/libcurl>).""", 'certain'),
                UpstreamDatum(
                    'X-Maintainer', Person('Scott Chamberlain', email='myrmecocystus@gmail.com'), 'certain'),
                UpstreamDatum(
                    "Repository", "https://github.com/ropensci/crul.git", "certain"
                ),
                UpstreamDatum("Homepage", "https://www.example.com/crul", "certain"),
            ],
        )


class GuessRepoFromUrlTests(TestCase):
    def test_github(self):
        self.assertEqual(
            "https://github.com/jelmer/blah",
            guess_repo_from_url("https://github.com/jelmer/blah"),
        )
        self.assertEqual(
            "https://github.com/jelmer/blah",
            guess_repo_from_url("https://github.com/jelmer/blah/blob/README"),
        )
        self.assertIs(None, guess_repo_from_url("https://github.com/jelmer"))

    def test_none(self):
        self.assertIs(None, guess_repo_from_url("https://www.jelmer.uk/"))

    def test_known(self):
        self.assertEqual(
            "http://code.launchpad.net/blah",
            guess_repo_from_url("http://code.launchpad.net/blah"),
        )

    def test_launchpad(self):
        self.assertEqual(
            "https://code.launchpad.net/bzr",
            guess_repo_from_url("http://launchpad.net/bzr/+download"),
        )

    def test_savannah(self):
        self.assertEqual(
            "https://git.savannah.gnu.org/git/auctex.git",
            guess_repo_from_url("https://git.savannah.gnu.org/git/auctex.git"),
        )
        self.assertIs(
            None, guess_repo_from_url("https://git.savannah.gnu.org/blah/auctex.git")
        )

    def test_bitbucket(self):
        self.assertEqual(
            "https://bitbucket.org/fenics-project/dolfin",
            guess_repo_from_url(
                "https://bitbucket.org/fenics-project/dolfin/downloads/"
            ),
        )


class BugDbFromBugSubmitUrlTests(TestCase):
    def test_github(self):
        self.assertEqual(
            "https://github.com/dulwich/dulwich/issues",
            bug_database_url_from_bug_submit_url(
                "https://github.com/dulwich/dulwich/issues/new"
            ),
        )

    def test_sf(self):
        self.assertEqual(
            "https://sourceforge.net/p/dulwich/bugs",
            bug_database_url_from_bug_submit_url(
                "https://sourceforge.net/p/dulwich/bugs/new"
            ),
        )


class UrlFromGitCloneTests(TestCase):
    def test_guess_simple(self):
        self.assertEqual(
            "https://github.com/jelmer/blah",
            url_from_git_clone_command(b"git clone https://github.com/jelmer/blah"),
        )
        self.assertEqual(
            "https://github.com/jelmer/blah",
            url_from_git_clone_command(
                b"git clone https://github.com/jelmer/blah target"
            ),
        )

    def test_args(self):
        self.assertEqual(
            "https://github.com/jelmer/blah",
            url_from_git_clone_command(
                b"git clone -b foo https://github.com/jelmer/blah target"
            ),
        )


class UrlFromFossilCloneTests(TestCase):
    def test_guess_simple(self):
        self.assertEqual(
            "https://example.com/repo/blah",
            url_from_fossil_clone_command(
                b"fossil clone https://example.com/repo/blah blah.fossil"
            ),
        )


class CertaintySufficientTests(TestCase):
    def test_sufficient(self):
        self.assertTrue(certainty_sufficient("certain", "certain"))
        self.assertTrue(certainty_sufficient("certain", "possible"))
        self.assertTrue(certainty_sufficient("certain", None))
        self.assertTrue(certainty_sufficient("possible", None))
        # TODO(jelmer): Should we really always allow unknown certainties
        # through?
        self.assertTrue(certainty_sufficient(None, "certain"))

    def test_insufficient(self):
        self.assertFalse(certainty_sufficient("possible", "certain"))


class CertaintyVsConfidenceTests(TestCase):
    def test_confidence_to_certainty(self):
        self.assertEqual("certain", confidence_to_certainty(0))
        self.assertEqual("confident", confidence_to_certainty(1))
        self.assertEqual("likely", confidence_to_certainty(2))
        self.assertEqual("possible", confidence_to_certainty(3))
        self.assertEqual("unknown", confidence_to_certainty(None))
        self.assertRaises(ValueError, confidence_to_certainty, 2000)

    def test_certainty_to_confidence(self):
        self.assertEqual(0, certainty_to_confidence("certain"))
        self.assertEqual(1, certainty_to_confidence("confident"))
        self.assertEqual(2, certainty_to_confidence("likely"))
        self.assertEqual(3, certainty_to_confidence("possible"))
        self.assertIs(None, certainty_to_confidence("unknown"))
        self.assertRaises(ValueError, certainty_to_confidence, "blah")


class MinimumCertaintyTests(TestCase):
    def test_minimum(self):
        self.assertEqual("certain", min_certainty([]))
        self.assertEqual("certain", min_certainty(["certain"]))
        self.assertEqual("possible", min_certainty(["possible"]))
        self.assertEqual("possible", min_certainty(["possible", "certain"]))
        self.assertEqual("likely", min_certainty(["likely", "certain"]))
        self.assertEqual("possible", min_certainty(["likely", "certain", "possible"]))



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Functions for working with upstream metadata.

This gathers information about upstreams from various places.
Each bit of information gathered is wrapped in a UpstreamDatum
object, which contains the field name.

The fields used here match those in https://wiki.debian.org/UpstreamMetadata

Supported fields:
- Homepage
- Name
- Contact
- Repository
- Repository-Browse
- Bug-Database
- Bug-Submit
- Screenshots
- Archive
- Security-Contact

Extensions for upstream-ontologist.
- X-SourceForge-Project: Name of the SourceForge project
- X-Wiki: URL to a wiki
- X-Summary: A one-line description
- X-Description: Multi-line description
- X-License: Short description of the license
- X-Copyright
- X-Maintainer
- X-Authors

Supported, but currently not set.
- FAQ
- Donation
- Documentation
- Registration
- Webservice
"""

import os
from typing import Optional, Sequence, TypeVar, Generic, List
from dataclasses import dataclass
from email.utils import parseaddr
from urllib.parse import urlparse


try:
    from typing import TypedDict  # type: ignore
except ImportError:
    from typing_extensions import TypedDict  # type: ignore


SUPPORTED_CERTAINTIES = ["certain", "confident", "likely", "possible", None]

version_string = "0.1.35"

USER_AGENT = "upstream-ontologist/" + version_string
# Too aggressive?
DEFAULT_URLLIB_TIMEOUT = 3


@dataclass
class Person:

    name: str
    email: Optional[str] = None
    url: Optional[str] = None

    def __init__(self, name, email=None, url=None):
        self.name = name
        self.email = email
        if url and url.startswith('mailto:'):
            self.email = url[len('mailto:'):]
            self.url = None
        else:
            self.url = url

    @classmethod
    def from_string(cls, text):
        text = text.replace(' at ', '@')
        text = text.replace(' -at- ', '@')
        text = text.replace(' -dot- ', '.')
        text = text.replace('[AT]', '@')
        if '(' in text and text.endswith(')'):
            (p1, p2) = text[:-1].split('(', 1)
            if p2.startswith('https://') or p2.startswith('http://'):
                url = p2
                if '<' in p1:
                    (name, email) = parseaddr(p1)
                    return cls(name=name, email=email, url=url)
                return cls(name=p1, url=url)
            elif '@' in p2:
                return cls(name=p1, email=p2)
            return cls(text)
        elif '<' in text:
            (name, email) = parseaddr(text)
            return cls(name=name, email=email)
        else:
            return cls(name=text)

    def __str__(self):
        if self.email:
            return '{} <{}>'.format(self.name, self.email)
        return self.name


T = TypeVar('T')


class UpstreamDatum(Generic[T]):
    """A single piece of upstream metadata."""

    __slots__ = ["field", "value", "certainty", "origin"]

    field: str
    value: T
    certainty: Optional[str]
    origin: Optional[str]

    def __init__(self, field: str, value: T, certainty: Optional[str] = None,
                 origin: Optional[str] = None) -> None:
        self.field = field
        if value is None:
            raise ValueError(field)
        self.value = value
        if certainty not in SUPPORTED_CERTAINTIES:
            raise ValueError(certainty)
        self.certainty = certainty
        self.origin = origin

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.field == other.field
            and self.value == other.value
            and self.certainty == other.certainty
            and self.origin == other.origin
        )

    def __str__(self):
        return "{}: {}".format(self.field, self.value)

    def __repr__(self):
        return "{}({!r}, {!r}, {!r}, {!r})".format(
            type(self).__name__,
            self.field,
            self.value,
            self.certainty,
            self.origin,
        )


UpstreamMetadata = TypedDict('UpstreamMetadata', {
    'Name': UpstreamDatum[str],
    'Contact': UpstreamDatum[str],
    'Repository': UpstreamDatum[str],
    'Repository-Browse': UpstreamDatum[str],
    'X-Summary': UpstreamDatum[str],
    'Bug-Database': UpstreamDatum[str],
    'Bug-Submit': UpstreamDatum[str],
    'Homepage': UpstreamDatum[str],
    'Screenshots': UpstreamDatum[List[str]],
}, total=False)


class UpstreamPackage:
    def __init__(self, family, name):
        self.family = family
        self.name = name


# If we're setting them new, put Name and Contact first
def upstream_metadata_sort_key(x):
    (k, v) = x
    return {
        "Name": "00-Name",
        "Contact": "01-Contact",
    }.get(k, k)


def min_certainty(certainties: Sequence[str]) -> str:
    confidences = [certainty_to_confidence(c) for c in certainties]
    return confidence_to_certainty(max([c for c in confidences if c is not None] + [0]))


def certainty_to_confidence(certainty: Optional[str]) -> Optional[int]:
    if certainty in ("unknown", None):
        return None
    return SUPPORTED_CERTAINTIES.index(certainty)


def confidence_to_certainty(confidence: Optional[int]) -> str:
    if confidence is None:
        return "unknown"
    try:
        return SUPPORTED_CERTAINTIES[confidence] or "unknown"
    except IndexError as e:
        raise ValueError(confidence) from e


def certainty_sufficient(
    actual_certainty: Optional[str], minimum_certainty: Optional[str]
) -> bool:
    """Check if the actual certainty is sufficient.

    Args:
      actual_certainty: Actual certainty with which changes were made
      minimum_certainty: Minimum certainty to keep changes
    Returns:
      boolean
    """
    actual_confidence = certainty_to_confidence(actual_certainty)
    if actual_confidence is None:
        # Actual confidence is unknown.
        # TODO(jelmer): Should we really be ignoring this?
        return True
    minimum_confidence = certainty_to_confidence(minimum_certainty)
    if minimum_confidence is None:
        return True
    return actual_confidence <= minimum_confidence


def _load_json_url(http_url: str, timeout: int = DEFAULT_URLLIB_TIMEOUT):
    from urllib.request import urlopen, Request
    import json
    headers = {
        'User-Agent': USER_AGENT,
        'Accept': 'application/json',
    }
    if urlparse(http_url).hostname in (
            'github.com', 'raw.githubusercontent.com'):
        try:
            headers['WWW-Authenticate'] = 'Bearer %s' % (
                os.environ['GITHUB_TOKEN'])
        except KeyError:
            pass
    http_contents = urlopen(
        Request(http_url, headers=headers),
        timeout=timeout).read()
    return json.loads(http_contents)


class UrlUnverifiable(Exception):
    """Unable to check specified URL."""

    def __init__(self, url, reason):
        self.url = url
        self.reason = reason


class InvalidUrl(Exception):
    """Specified URL is invalid."""

    def __init__(self, url, reason):
        self.url = url
        self.reason = reason



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""README parsing."""

import logging
import platform
import re
from typing import Optional, Tuple, Iterable, List
from urllib.parse import urlparse

from . import UpstreamDatum


logger = logging.getLogger(__name__)


def _skip_paragraph(para, metadata):  # noqa: C901
    if re.match(r'See .* for more (details|information)\.', para):
        return True
    if re.match(r'See .* for instructions', para):
        return True
    if re.match(r'Please refer .*\.', para):
        return True
    m = re.match(r'It is licensed under (.*)', para)
    if m:
        metadata.append(UpstreamDatum('X-License', m.group(1), 'possible'))
        return True
    m = re.match(r'License: (.*)', para, re.I)
    if m:
        metadata.append(UpstreamDatum('X-License', m.group(1), 'likely'))
        return True
    m = re.match('(Home page|homepage_url|Main website|Website|Homepage): (.*)', para, re.I)
    if m:
        url = m.group(2)
        if url.startswith('<') and url.endswith('>'):
            url = url[1:-1]
        metadata.append(UpstreamDatum('Homepage', url, 'likely'))
        return True
    m = re.match('More documentation .* at http.*', para)
    if m:
        return True
    m = re.match('Documentation (can be found|is hosted|is available) (at|on) ([^ ]+)', para)
    if m:
        metadata.append(UpstreamDatum('Documentation', m.group(3), 'likely'))
        return True
    m = re.match(r'Documentation for (.*)\s+(can\s+be\s+found|is\s+hosted)\s+(at|on)\s+([^ ]+)', para)
    if m:
        metadata.append(UpstreamDatum('Name', m.group(1), 'possible'))
        metadata.append(UpstreamDatum('Documentation', m.group(4), 'likely'))
        return True
    if re.match(r'Documentation[, ].*found.*(at|on).*\.', para, re.S):
        return True
    m = re.match('See (http.*|gopkg.in.*|github.com.*)', para)
    if m:
        return True
    m = re.match('Available on (.*)', para)
    if m:
        return True
    m = re.match(
        r'This software is freely distributable under the (.*) license.*',
        para)
    if m:
        metadata.append(UpstreamDatum('X-License', m.group(1), 'likely'))
        return True
    m = re.match(r'This .* is hosted at .*', para)
    if m:
        return True
    m = re.match(r'This code has been developed by .*', para)
    if m:
        return True
    if para.startswith('Download and install using:'):
        return True
    m = re.match('Bugs should be reported by .*', para)
    if m:
        return True
    m = re.match(r'The bug tracker can be found at (http[^ ]+[^.])', para)
    if m:
        metadata.append(UpstreamDatum('Bug-Database', m.group(1), 'likely'))
        return True
    m = re.match(r'Copyright (\(c\) |)(.*)', para)
    if m:
        metadata.append(UpstreamDatum('X-Copyright', m.group(2), 'possible'))
        return True
    if re.match('You install .*', para):
        return True
    if re.match('This .* is free software; .*', para):
        return True
    m = re.match('Please report any bugs(.*) to <(.*)>', para)
    if m:
        metadata.append(UpstreamDatum('Bugs-Submit', m.group(2), 'possible'))
        return True
    if re.match('Share and Enjoy', para, re.I):
        return True
    lines = para.splitlines(False)
    if lines and lines[0].strip() in ('perl Makefile.PL', 'make', './configure'):
        return True
    if re.match('For further information, .*', para):
        return True
    if re.match('Further information .*', para):
        return True
    m = re.match(r'A detailed Changelog can be found.*:\s+(http.*)', para, re.I)
    if m:
        metadata.append(UpstreamDatum('Changelog', m.group(1), 'possible'))
        return True


def _skip_paragraph_block(para, metadata):  # noqa: C901
    if _skip_paragraph(para.get_text(), metadata):
        return True
    for c in para.children:
        if isinstance(c, str) and not c.strip():
            continue
        if c.name == 'a':
            if len(list(c.children)) != 1:
                name = None
            elif isinstance(list(c.children)[0], str):
                name = list(c.children)[0]
            elif list(c.children)[0].name == 'img':
                name = list(c.children)[0].get('alt')
            else:
                name = None
            if name in ('CRAN', 'CRAN_Status_Badge', 'CRAN_Logs_Badge'):
                metadata.append(UpstreamDatum('Archive', 'CRAN', 'confident'))
            elif name == 'Gitter':
                parsed_url = urlparse(c.get('href'))
                metadata.append(UpstreamDatum(
                    'Repository',
                    'https://github.com/%s' % '/'.join(parsed_url.path.strip('/').split('/')[:2]),
                    'confident'))
            elif name and name.lower() == 'build status':
                parsed_url = urlparse(c.get('href'))
                if parsed_url.hostname == 'travis-ci.org':
                    metadata.append(UpstreamDatum(
                        'Repository',
                        'https://github.com/%s' % '/'.join(parsed_url.path.strip('/').split('/')[:2]),
                        'confident'))
            elif name and name.lower() == 'documentation':
                metadata.append(UpstreamDatum(
                    'Documentation', c.get('href'), 'confident'))
            elif name and name.lower() == 'api docs':
                metadata.append(UpstreamDatum(
                    'X-API-Documentation', c.get('href'), 'confident'))
            elif name and name.lower() == 'downloads':
                metadata.append(UpstreamDatum(
                    'X-Download', c.get('href'), 'confident'))
            elif name and name.lower() == 'crates.io':
                href = c.get('href')
                if href.startswith('https://crates.io/crates/'):
                    metadata.append(UpstreamDatum(
                        'X-Cargo-Crate', href.rsplit('/')[-1], 'confident'))
            elif name:
                m = re.match('(.*) License', name)
                if m:
                    metadata.append(UpstreamDatum('X-License', m.group(1), 'likely'))
                else:
                    logger.debug('Unhandled field %r in README', name)
            continue
        break
    else:
        return True
    if para.get_text() == '':
        return True
    return False


def render(el):
    return el.get_text()


def _parse_first_header_text(text):
    m = re.fullmatch('([A-Za-z]+) ([0-9.]+)', text)
    if m:
        return m.group(1), None, m.group(2)
    m = re.fullmatch('([A-Za-z]+): (.+)', text)
    if m:
        return m.group(1), m.group(2), None
    m = re.fullmatch('([A-Za-z]+) - (.+)', text)
    if m:
        return m.group(1), m.group(2), None
    m = re.fullmatch('([A-Za-z]+) -- (.+)', text)
    if m:
        return m.group(1), m.group(2), None
    m = re.fullmatch('([A-Za-z]+) version ([^ ]+)', text)
    if m:
        name, version = text.split(' version ', 1)
        summary = None
        return name, summary, version
    return None, None, None


def _parse_first_header(el):
    name, summary, version = _parse_first_header_text(el.get_text())
    if not name and el.get_text():
        name = el.get_text()
    if name:
        if 'installation' in name.lower():
            certainty = 'possible'
        else:
            certainty = 'likely'
        if name.startswith('About '):
            name = name[len('About '):]
        yield UpstreamDatum('Name', name.strip(), certainty)
    if summary:
        yield UpstreamDatum('X-Summary', summary, 'likely')
    if version:
        yield UpstreamDatum('X-Version', version, 'likely')


def _is_semi_header(el):
    if el.name != 'p':
        return False
    if el.get_text().strip() == 'INSTALLATION':
        return True
    if el.get_text().count('\n') > 0:
        return False
    m = re.match(r'([a-z-A-Z0-9]+) - ([^\.]+)', el.get_text())
    if m:
        return True
    return False


def _ul_is_field_list(el):
    names = ['Issues', 'Home', 'Documentation', 'License']
    for li in el.findAll('li'):
        m = re.match(r'([A-Za-z]+)\s*:.*', li.get_text().strip())
        if not m or m.group(1) not in names:
            return False
    return True


def _extract_paragraphs(children, metadata):
    paragraphs = []
    for el in children:
        if isinstance(el, str):
            continue
        if el.name == 'div':
            paragraphs.extend(_extract_paragraphs(el.children, metadata))
            if paragraphs and 'section' in (el.get('class') or []):
                break
        if el.name == 'p':
            if _is_semi_header(el):
                if len(paragraphs) == 0:
                    metadata.extend(_parse_first_header(el))
                    continue
                else:
                    break
            if _skip_paragraph_block(el, metadata):
                if len(paragraphs) > 0:
                    break
                else:
                    continue
            if el.get_text().strip():
                paragraphs.append(render(el) + '\n')
        elif el.name == 'pre':
            paragraphs.append(render(el))
        elif el.name == 'ul' and len(paragraphs) > 0:
            if _ul_is_field_list(el):
                metadata.extend(_parse_ul_field_list(el))
            else:
                paragraphs.append(
                    ''.join(
                        '* %s\n' % li.get_text()
                        for li in el.findAll('li')))
        elif re.match('h[0-9]', el.name):
            if len(paragraphs) == 0:
                if el.get_text() not in ('About', 'Introduction', 'Overview'):
                    metadata.extend(_parse_first_header(el))
                continue
            break
    return paragraphs


def _parse_field(name, body):
    if name == 'Homepage' and body.find('a'):
        yield UpstreamDatum('Homepage', body.find('a').get('href'), 'confident')
    if name == 'Home' and body.find('a'):
        yield UpstreamDatum('Homepage', body.find('a').get('href'), 'confident')
    if name == 'Issues' and body.find('a'):
        yield UpstreamDatum('Bug-Database', body.find('a').get('href'), 'confident')
    if name == 'Documentation' and body.find('a'):
        yield UpstreamDatum('Documentation', body.find('a').get('href'), 'confident')
    if name == 'License':
        yield UpstreamDatum('X-License', body.get_text(), 'confident')


def _parse_ul_field_list(el):
    for li in el.findAll('li'):
        cs = list(li.children)
        if len(cs) == 2 and isinstance(cs[0], str):
            name = cs[0].strip().rstrip(':')
            body = cs[1]
            yield from _parse_field(name, body)


def _parse_field_list(tab):
    for tr in tab.findAll('tr', {'class': 'field'}):
        name_cell = tr.find('th', {'class': 'field-name'})
        if not name_cell:
            continue
        name = name_cell.get_text().rstrip(':')
        body = tr.find('td', {'class': 'field-body'})
        if not body:
            continue
        yield from _parse_field(name, body)


def _description_from_basic_soup(soup) -> Tuple[Optional[str], Iterable[UpstreamDatum]]:
    # Drop any headers
    metadata = []
    if soup is None:
        return None, {}
    # First, skip past the first header.
    for el in soup.children:
        if el.name in ('h1', 'h2', 'h3'):
            metadata.extend(_parse_first_header(el))
            el.decompose()
            break
        elif isinstance(el, str):
            pass
        else:
            break

    table = soup.find('table', {'class': 'field-list'})
    if table:
        metadata.extend(_parse_field_list(table))

    paragraphs: List[str] = []
    paragraphs.extend(_extract_paragraphs(soup.children, metadata))

    if len(paragraphs) == 0:
        logger.debug('Empty description; no paragraphs.')
        return None, metadata

    if len(paragraphs) < 6:
        return '\n'.join(paragraphs), metadata
    logger.debug(
        'Not returning description, number of paragraphs too high: %d',
        len(paragraphs))
    return None, metadata


def description_from_readme_md(md_text: str) -> Tuple[Optional[str], Iterable[UpstreamDatum]]:
    """Description from README.md."""
    try:
        import markdown
    except ModuleNotFoundError:
        logger.debug('markdown not available, not parsing README.md')
        return None, {}
    html_text = markdown.markdown(md_text)
    try:
        from bs4 import BeautifulSoup, FeatureNotFound
    except ModuleNotFoundError:
        logger.debug('BeautifulSoup not available, not parsing README.md')
        return None, {}
    # Strip surrogates
    html_text = html_text.encode('utf-8', 'replace').decode('utf-8')
    try:
        soup = BeautifulSoup(html_text, 'lxml')
    except FeatureNotFound:
        logger.debug('lxml not available, not parsing README.md')
        return None, {}
    return _description_from_basic_soup(soup.body)


def description_from_readme_rst(rst_text: str) -> Tuple[Optional[str], Iterable[UpstreamDatum]]:
    """Description from README.rst."""
    if platform.python_implementation() == "PyPy":
        logger.debug('docutils does not appear to work on PyPy, skipping README.rst.')
        return None, {}
    try:
        from docutils.core import publish_parts
    except ModuleNotFoundError:
        logger.debug('docutils not available, not parsing README.rst')
        return None, {}

    from docutils.writers.html4css1 import Writer
    settings = {'initial_header_level': 2, 'report_level': 0}
    html_text = publish_parts(
        rst_text, writer=Writer(), settings_overrides=settings).get('html_body')
    try:
        from bs4 import BeautifulSoup, FeatureNotFound
    except ModuleNotFoundError:
        logger.debug('BeautifulSoup not available, not parsing README.rst')
        return None, {}
    try:
        soup = BeautifulSoup(html_text, 'lxml')
    except FeatureNotFound:
        logger.debug('lxml not available, not parsing README.rst')
        return None, {}
    if not soup.body:
        return None, {}
    return _description_from_basic_soup(list(soup.body.children)[0])


def description_from_readme_plain(text: str) -> Tuple[Optional[str], Iterable[UpstreamDatum]]:
    lines = list(text.splitlines(False))
    metadata = []
    if not lines:
        return None, {}
    if lines[0].strip() and len(lines) > 1 and (not lines[1] or not lines[1][0].isalnum()):
        name, summary, version = _parse_first_header_text(lines[0])
        if name:
            metadata.append(UpstreamDatum('Name', name, 'likely'))
        if version:
            metadata.append(UpstreamDatum('X-Version', version, 'likely'))
        if summary:
            metadata.append(UpstreamDatum('X-Summary', summary, 'likely'))
        if name or version or summary:
            lines.pop(0)
    else:
        name = version = summary = None
    while lines and not lines[0].strip('-').strip():
        lines.pop(0)

    paras: List[List[str]] = [[]]
    for line in lines:
        if not line.strip():
            paras.append([])
        else:
            paras[-1].append(line)

    output: List[str] = []
    for para in paras:
        if not para:
            continue
        line = '\n'.join(para)
        if _skip_paragraph(line, metadata):
            continue
        output.append(line + '\n')
    if len(output) > 30:
        return None, {}
    while output and not output[-1].strip():
        output.pop(-1)
    return '\n'.join(output), metadata



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import re
from typing import Optional

from .. import UpstreamPackage


def debian_to_upstream_version(version):
    """Drop debian-specific modifiers from an upstream version string."""
    return version.upstream_version.split("+dfsg")[0]


def upstream_name_to_debian_source_name(upstream_name: str) -> str:
    m = re.match(r'^(.{10,})\((.*)\)', upstream_name)
    if m:
        upstream_name = m.group(2)
    if upstream_name.startswith("GNU "):
        upstream_name = upstream_name[len("GNU ") :]
    return upstream_name.lower().replace('_', '-').replace(' ', '-').replace('/', '-')


def upstream_version_to_debian_upstream_version(
    version: str, family: Optional[str] = None
) -> str:
    # TODO(jelmer)
    return version


def upstream_package_to_debian_source_name(package: UpstreamPackage) -> str:
    if package.family == "rust":
        return "rust-%s" % package.name.lower()
    if package.family == "perl":
        return "lib%s-perl" % package.name.lower().replace("::", "-")
    if package.family == "node":
        return "node-%s" % package.name.lower()
    # TODO(jelmer):
    return upstream_name_to_debian_source_name(package.name)


def upstream_package_to_debian_binary_name(package: UpstreamPackage) -> str:
    if package.family == "rust":
        return "rust-%s" % package.name.lower()
    if package.family == "perl":
        return "lib%s-perl" % package.name.lower().replace("::", "-")
    if package.family == "node":
        return "node-%s" % package.name.lower()
    # TODO(jelmer):
    return package.name.lower().replace('_', '-')


def compare_upstream_versions(family, version1, version2):
    raise NotImplementedError


package_name_re = re.compile("[a-z0-9][a-z0-9+-.]+")


def valid_debian_package_name(name):
    return bool(package_name_re.fullmatch(name))
